@@ -2,23 +2,21 @@
 # this repository contains the full copyright notices and license terms.
 
 import datetime
-from dateutil.relativedelta import relativedelta
 import itertools
 
+from dateutil.relativedelta import relativedelta
 from sql import Null
 from sql.aggregate import Sum
 from sql.conditionals import Coalesce
 
 from trytond.i18n import gettext
-from trytond.model import ModelView, Workflow, ModelSQL, fields, Unique
+from trytond.model import Index, ModelSQL, ModelView, Unique, Workflow, fields
 from trytond.model.exceptions import AccessError
-from trytond.wizard import Wizard, StateView, StateTransition, Button
-from trytond.pyson import Not, Equal, Eval, Or, Bool, If
-from trytond.transaction import Transaction
 from trytond.pool import Pool
-from trytond.tools import reduce_ids, grouped_slice
-
-from trytond.modules.product import round_price
+from trytond.pyson import Bool, Equal, Eval, If, Not, Or
+from trytond.tools import grouped_slice, reduce_ids
+from trytond.transaction import Transaction
+from trytond.wizard import Button, StateTransition, StateView, Wizard
 
 from .exceptions import ForecastValidationError
 
@@ -30,49 +28,50 @@ class Forecast(Workflow, ModelSQL, ModelView):
     _states = {
         'readonly': Not(Equal(Eval('state'), 'draft')),
     }
-    _depends = ['state']
 
     warehouse = fields.Many2One(
         'stock.location', 'Location', required=True,
         domain=[('type', '=', 'warehouse')], states={
             'readonly': Or(Not(Equal(Eval('state'), 'draft')),
                 Bool(Eval('lines', [0]))),
-            },
-        depends=['state'])
+            })
     destination = fields.Many2One(
         'stock.location', 'Destination', required=True,
-        domain=[('type', 'in', ['customer', 'production'])], states=_states,
-        depends=_depends)
+        domain=[('type', 'in', ['customer', 'production'])], states=_states)
     from_date = fields.Date(
         "From Date", required=True,
         domain=[('from_date', '<=', Eval('to_date'))],
-        states=_states, depends=_depends + ['to_date'])
+        states=_states)
     to_date = fields.Date(
         "To Date", required=True,
         domain=[('to_date', '>=', Eval('from_date'))],
-        states=_states, depends=_depends + ['from_date'])
+        states=_states)
     lines = fields.One2Many(
-        'stock.forecast.line', 'forecast', 'Lines', states=_states,
-        depends=_depends)
+        'stock.forecast.line', 'forecast', 'Lines', states=_states)
     company = fields.Many2One(
         'company.company', 'Company', required=True, states={
             'readonly': Or(Not(Equal(Eval('state'), 'draft')),
                 Bool(Eval('lines', [0]))),
-            },
-        depends=['state'])
+            })
     state = fields.Selection([
             ('draft', "Draft"),
             ('done', "Done"),
             ('cancelled', "Cancelled"),
-            ], "State", readonly=True, select=True)
+            ], "State", readonly=True, sort=False)
     active = fields.Function(fields.Boolean('Active'),
         'get_active', searcher='search_active')
 
-    del _states, _depends
+    del _states
 
     @classmethod
     def __setup__(cls):
         super(Forecast, cls).__setup__()
+        t = cls.__table__()
+        cls._sql_indexes.add(
+            Index(
+                t,
+                (t.state, Index.Equality()),
+                (t.to_date, Index.Range())))
         cls.create_date.select = True
         cls._order.insert(0, ('from_date', 'DESC'))
         cls._order.insert(1, ('warehouse', 'ASC'))
@@ -162,44 +161,66 @@ class Forecast(Workflow, ModelSQL, ModelView):
             return []
 
     def get_rec_name(self, name):
-        return self.warehouse.rec_name
+        pool = Pool()
+        Lang = pool.get('ir.lang')
+
+        lang = Lang.get()
+        from_date = lang.strftime(self.from_date)
+        to_date = lang.strftime(self.to_date)
+        return (
+            f'{self.warehouse.rec_name} → {self.destination.rec_name} @ '
+            f'[{from_date} - {to_date}]')
 
     @classmethod
     def search_rec_name(cls, name, clause):
-        return [('warehouse.rec_name',) + tuple(clause[1:])]
+        operator = clause[1]
+        if operator.startswith('!') or operator.startswith('not '):
+            bool_op = 'AND'
+        else:
+            bool_op = 'OR'
+        return [bool_op,
+            ('warehouse.rec_name', *clause[1:]),
+            ('destination.rec_name', *clause[1:]),
+            ]
 
     @classmethod
-    def validate(cls, forecasts):
-        super(Forecast, cls).validate(forecasts)
-        for forecast in forecasts:
-            forecast.check_date_overlap()
+    def validate_fields(cls, forecasts, field_names):
+        super().validate_fields(forecasts, field_names)
+        cls.check_date_overlap(forecasts, field_names)
 
-    def check_date_overlap(self):
-        if self.state != 'done':
+    @classmethod
+    def check_date_overlap(cls, forecasts, field_names=None):
+        if field_names and not (field_names & {
+                    'from_date', 'to_date',
+                    'warehouse', 'destination',
+                    'company', 'state'}):
             return
         transaction = Transaction()
         connection = transaction.connection
-        transaction.database.lock(connection, self._table)
-        forcast = self.__table__()
+        cls.lock()
+        table = cls.__table__()
         cursor = connection.cursor()
-        cursor.execute(*forcast.select(forcast.id,
-                where=(((forcast.from_date <= self.from_date)
-                        & (forcast.to_date >= self.from_date))
-                    | ((forcast.from_date <= self.to_date)
-                        & (forcast.to_date >= self.to_date))
-                    | ((forcast.from_date >= self.from_date)
-                        & (forcast.to_date <= self.to_date)))
-                & (forcast.warehouse == self.warehouse.id)
-                & (forcast.destination == self.destination.id)
-                & (forcast.company == self.company.id)
-                & (forcast.id != self.id)))
-        forecast_id = cursor.fetchone()
-        if forecast_id:
-            second = self.__class__(forecast_id[0])
-            raise ForecastValidationError(
-                gettext('stock_forecast.msg_forecast_date_overlap',
-                    first=self.rec_name,
-                    second=second.rec_name))
+        for forecast in forecasts:
+            if forecast.state != 'done':
+                continue
+            cursor.execute(*table.select(table.id,
+                    where=(((table.from_date <= forecast.from_date)
+                            & (table.to_date >= forecast.from_date))
+                        | ((table.from_date <= forecast.to_date)
+                            & (table.to_date >= forecast.to_date))
+                        | ((table.from_date >= forecast.from_date)
+                            & (table.to_date <= forecast.to_date)))
+                    & (table.warehouse == forecast.warehouse.id)
+                    & (table.destination == forecast.destination.id)
+                    & (table.company == forecast.company.id)
+                    & (table.id != forecast.id)))
+            forecast_id = cursor.fetchone()
+            if forecast_id:
+                second = cls(forecast_id[0])
+                raise ForecastValidationError(
+                    gettext('stock_forecast.msg_forecast_date_overlap',
+                        first=forecast.rec_name,
+                        second=second.rec_name))
 
     @classmethod
     def delete(cls, forecasts):
@@ -261,35 +282,35 @@ class ForecastLine(ModelSQL, ModelView):
     _states = {
         'readonly': Eval('forecast_state') != 'draft',
         }
-    _depends = ['forecast_state']
 
     product = fields.Many2One('product.product', 'Product', required=True,
         domain=[
             ('type', '=', 'goods'),
             ('consumable', '=', False),
             ],
-        states=_states, depends=_depends)
+        states=_states)
     product_uom_category = fields.Function(
-        fields.Many2One('product.uom.category', 'Product Uom Category'),
+        fields.Many2One(
+            'product.uom.category', "Product UoM Category",
+            help="The category of Unit of Measure for the product."),
         'on_change_with_product_uom_category')
-    uom = fields.Many2One('product.uom', 'UOM', required=True,
+    unit = fields.Many2One(
+        'product.uom', "Unit", required=True,
         domain=[
             If(Bool(Eval('product_uom_category')),
                 ('category', '=', Eval('product_uom_category')),
                 ('category', '!=', -1)),
             ],
         states=_states,
-        depends=['product', 'product_uom_category'] + _depends)
-    unit_digits = fields.Function(fields.Integer('Unit Digits'),
-            'get_unit_digits')
+        depends={'product'})
     quantity = fields.Float(
-        "Quantity", digits=(16, Eval('unit_digits', 2)), required=True,
+        "Quantity", digits='unit', required=True,
         domain=[('quantity', '>=', 0)],
-        states=_states, depends=['unit_digits'] + _depends)
+        states=_states)
     minimal_quantity = fields.Float(
-        "Minimal Qty", digits=(16, Eval('unit_digits', 2)), required=True,
+        "Minimal Qty", digits='unit', required=True,
         domain=[('minimal_quantity', '<=', Eval('quantity'))],
-        states=_states, depends=['unit_digits', 'quantity'] + _depends)
+        states=_states)
     moves = fields.Many2Many('stock.forecast.line-stock.move',
         'line', 'move', 'Moves', readonly=True)
     forecast = fields.Many2One(
@@ -297,16 +318,14 @@ class ForecastLine(ModelSQL, ModelView):
         states={
             'readonly': ((Eval('forecast_state') != 'draft')
                 & Bool(Eval('forecast'))),
-            },
-        depends=['forecast_state'])
+            })
     forecast_state = fields.Function(
         fields.Selection('get_forecast_states', 'Forecast State'),
         'on_change_with_forecast_state')
-    quantity_executed = fields.Function(fields.Float('Quantity Executed',
-            digits=(16, Eval('unit_digits', 2)), depends=['unit_digits']),
-        'get_quantity_executed')
+    quantity_executed = fields.Function(fields.Float(
+            "Quantity Executed", digits='unit'), 'get_quantity_executed')
 
-    del _states, _depends
+    del _states
 
     @classmethod
     def __setup__(cls):
@@ -320,6 +339,13 @@ class ForecastLine(ModelSQL, ModelView):
 
     @classmethod
     def __register__(cls, module_name):
+        table_h = cls.__table_handler__(module_name)
+
+        # Migration from 6.8: rename uom to unit
+        if (table_h.column_exist('uom')
+                and not table_h.column_exist('unit')):
+            table_h.column_rename('uom', 'unit')
+
         super().__register__(module_name)
 
         table_h = cls.__table_handler__(module_name)
@@ -328,34 +354,18 @@ class ForecastLine(ModelSQL, ModelView):
         table_h.drop_constraint('check_line_qty_pos')
         table_h.drop_constraint('check_line_minimal_qty')
 
-    @staticmethod
-    def default_unit_digits():
-        return 2
-
     @staticmethod
     def default_minimal_quantity():
         return 1.0
 
     @fields.depends('product')
     def on_change_product(self):
-        self.unit_digits = 2
         if self.product:
-            self.uom = self.product.default_uom
-            self.unit_digits = self.product.default_uom.digits
+            self.unit = self.product.default_uom
 
     @fields.depends('product')
     def on_change_with_product_uom_category(self, name=None):
-        if self.product:
-            return self.product.default_uom_category.id
-
-    @fields.depends('uom')
-    def on_change_uom(self):
-        self.unit_digits = 2
-        if self.uom:
-            self.unit_digits = self.uom.digits
-
-    def get_unit_digits(self, name):
-        return self.product.default_uom.digits
+        return self.product.default_uom_category if self.product else None
 
     @classmethod
     def get_forecast_states(cls):
@@ -422,8 +432,8 @@ class ForecastLine(ModelSQL, ModelView):
                         group_by=move.product))
                 for product_id, quantity in cursor:
                     line = product2line[product_id]
-                    result[line.id] = Uom.compute_qty(line.product.default_uom,
-                        quantity, line.uom)
+                    result[line.id] = Uom.compute_qty(
+                        line.product.default_uom, quantity, line.unit)
         return result
 
     @classmethod
@@ -439,7 +449,6 @@ class ForecastLine(ModelSQL, ModelView):
         'Get stock moves for the forecast line'
         pool = Pool()
         Move = pool.get('stock.move')
-        Uom = pool.get('product.uom')
         Date = pool.get('ir.date')
 
         assert not self.moves
@@ -451,18 +460,14 @@ class ForecastLine(ModelSQL, ModelView):
         to_date = self.forecast.to_date
         if to_date < today:
             return []
+        if self.quantity_executed >= self.quantity:
+            return []
 
         delta = to_date - from_date
         delta = delta.days + 1
         nb_packet = ((self.quantity - self.quantity_executed)
             // self.minimal_quantity)
         distribution = self.distribute(delta, nb_packet)
-        unit_price = None
-        if self.forecast.destination.type == 'customer':
-            unit_price = self.product.list_price
-            unit_price = Uom.compute_price(self.product.default_uom,
-                unit_price, self.uom)
-            unit_price = round_price(unit_price)
 
         moves = []
         for day, qty in distribution.items():
@@ -472,12 +477,13 @@ class ForecastLine(ModelSQL, ModelView):
             move.from_location = self.forecast.warehouse.storage_location
             move.to_location = self.forecast.destination
             move.product = self.product
-            move.uom = self.uom
+            move.unit = self.unit
             move.quantity = qty * self.minimal_quantity
             move.planned_date = from_date + datetime.timedelta(day)
             move.company = self.forecast.company
             move.currency = self.forecast.company.currency
-            move.unit_price = unit_price
+            move.unit_price = (
+                0 if self.forecast.destination.type == 'customer' else None)
             moves.append(move)
         return moves
 
@@ -519,34 +525,49 @@ class ForecastLineMove(ModelSQL):
     'ForecastLine - Move'
     __name__ = 'stock.forecast.line-stock.move'
     _table = 'forecast_line_stock_move_rel'
-    line = fields.Many2One('stock.forecast.line', 'Forecast Line',
-            ondelete='CASCADE', select=True, required=True)
-    move = fields.Many2One('stock.move', 'Move', ondelete='CASCADE',
-            select=True, required=True)
+    line = fields.Many2One(
+        'stock.forecast.line', "Forecast Line",
+        ondelete='CASCADE', required=True)
+    move = fields.Many2One(
+        'stock.move', "Move", ondelete='CASCADE', required=True)
 
 
 class ForecastCompleteAsk(ModelView):
     'Complete Forecast'
     __name__ = 'stock.forecast.complete.ask'
+    company = fields.Many2One('company.company', "Company", readonly=True)
+    warehouse = fields.Many2One('stock.location', "Warehouse", readonly=True)
+    destination = fields.Many2One(
+        'stock.location', "Destination", readonly=True)
     from_date = fields.Date(
         "From Date", required=True,
         domain=[('from_date', '<', Eval('to_date'))],
-        depends=['to_date'])
+        states={
+            'readonly': Bool(Eval('products')),
+            })
     to_date = fields.Date(
         "To Date", required=True,
         domain=[('to_date', '>', Eval('from_date'))],
-        depends=['from_date'])
-
-
-class ForecastCompleteChoose(ModelView):
-    'Complete Forecast'
-    __name__ = 'stock.forecast.complete.choose'
+        states={
+            'readonly': Bool(Eval('products')),
+            })
     products = fields.Many2Many(
         'product.product', None, None, "Products",
         domain=[
             ('type', '=', 'goods'),
             ('consumable', '=', False),
-            ])
+            ],
+        context={
+            'company': Eval('company', -1),
+            'locations': [Eval('warehouse', -1)],
+            'stock_destinations': [Eval('destination', -1)],
+            'stock_date_start': Eval('from_date', None),
+            'stock_date_end': Eval('to_date', None),
+            'with_childs': True,
+            'stock_invert': True,
+            },
+        depends=[
+            'company', 'warehouse', 'destination', 'from_date', 'to_date'])
 
 
 class ForecastComplete(Wizard):
@@ -555,15 +576,8 @@ class ForecastComplete(Wizard):
     start_state = 'ask'
     ask = StateView('stock.forecast.complete.ask',
         'stock_forecast.forecast_complete_ask_view_form', [
-            Button('Cancel', 'end', 'tryton-cancel'),
-            Button('Choose Products', 'choose', 'tryton-forward'),
-            Button('Complete', 'complete', 'tryton-ok', default=True),
-            ])
-    choose = StateView('stock.forecast.complete.choose',
-        'stock_forecast.forecast_complete_choose_view_form', [
-            Button('Cancel', 'end', 'tryton-cancel'),
-            Button('Choose Dates', 'ask', 'tryton-back'),
-            Button('Complete', 'complete', 'tryton-ok', default=True),
+            Button("Cancel", 'end', 'tryton-cancel'),
+            Button("Complete", 'complete', 'tryton-ok', default=True),
             ])
     complete = StateTransition()
 
@@ -572,74 +586,35 @@ class ForecastComplete(Wizard):
         Forecast dates shifted by one year.
         """
         default = {}
-        for field in ("to_date", "from_date"):
-            default[field] = (
-                getattr(self.record, field) - relativedelta(years=1))
+        for field in ['company', 'warehouse', 'destination']:
+            if field in fields:
+                record = getattr(self.record, field)
+                default[field] = record.id
+        for field in ["to_date", "from_date"]:
+            if field in fields:
+                default[field] = (
+                    getattr(self.record, field) - relativedelta(years=1))
         return default
 
-    def _get_product_quantity(self):
-        pool = Pool()
-        Product = pool.get('product.product')
-
-        with Transaction().set_context(
-                stock_destinations=[self.record.destination.id],
-                stock_date_start=self.ask.from_date,
-                stock_date_end=self.ask.to_date):
-            return Product.products_by_location([self.record.warehouse.id],
-                with_childs=True)
-
-    def default_choose(self, fields):
-        """
-        Collect products for which there is an outgoing stream between
-        the given location and the destination.
-        """
-        if getattr(self.choose, 'products', None):
-            return {'products': [x.id for x in self.choose.products]}
-        pbl = self._get_product_quantity()
-        products = []
-        for (_, product), qty in pbl.items():
-            if qty < 0:
-                products.append(product)
-        return {'products': products}
-
     def transition_complete(self):
         pool = Pool()
         ForecastLine = pool.get('stock.forecast.line')
-        Product = pool.get('product.product')
-
-        forecast = self.record
-        prod2line = {}
-        forecast_lines = ForecastLine.search([
-                ('forecast', '=', forecast.id),
-                ])
-        for forecast_line in forecast_lines:
-            prod2line[forecast_line.product.id] = forecast_line
-
-        pbl = self._get_product_quantity()
-        id2product = {p.id: p for p in Product.browse([x[1] for x in pbl])}
-
-        products = set(getattr(self.choose, 'products', {}))
 
+        product2line = {l.product: l for l in self.record.lines}
         to_save = []
-        for key, qty in pbl.items():
-            _, product_id = key
-            product = id2product[product_id]
-            if products and product not in products:
-                continue
-            if product.type != 'goods' or product.consumable:
-                continue
-            if -qty <= 0:
-                continue
-            if product in prod2line:
-                line = prod2line[product]
-            else:
-                line = ForecastLine()
-            line.product = product
-            line.quantity = -qty
-            line.uom = product.default_uom.id
-            line.forecast = forecast
-            line.minimal_quantity = min(1, -qty)
+        # Ensure context is set
+        self.ask.products = map(int, self.ask.products)
+        for product in self.ask.products:
+            line = product2line.get(product, ForecastLine())
+            self._fill_line(line, product)
             to_save.append(line)
-
         ForecastLine.save(to_save)
         return 'end'
+
+    def _fill_line(self, line, product):
+        quantity = max(product.quantity, 0)
+        line.product = product
+        line.quantity = quantity
+        line.unit = product.default_uom
+        line.forecast = self.record
+        line.minimal_quantity = min(1, quantity)
